#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Date    : 2014-10-06 17:19:06
# @Author  : Han Zhao (han.zhao@uwaterloo.ca)
# @Link    : https://github.com/KeiraZhao
# @Version : $Id$
import os, sys
import unittest
import theano
import theano.tensor as T
import numpy as np
import scipy as sp
import logging
import time
import csv
import traceback
import random
import argparse

from sklearn.cross_validation import KFold
from scipy.sparse import *

sys.path.append('../source/')

from grcnn import AdaSent
from config import AdaSentConfiger
from wordvec import WordEmbedding
from utils import floatX

charas = 'ABCDEFGHIJKLMNOPQRSTUVWXYZ'
default_name = ''.join([charas[np.random.randint(0, len(charas))] for _ in xrange(5)])

parser = argparse.ArgumentParser()
parser.add_argument('-s', '--size', help='The size of each batch used to be trained.', 
                    type=int, default=20)
parser.add_argument('-l', '--rate', help='Learning rate of AdaGrad.', 
                    type=float, default=0.02)
parser.add_argument('-e', '--epoch', help='Number of the training epoch.',
                    type=int, default=50)
parser.add_argument('-n', '--name', help='Name used to save the model.', 
                    type=str, default=default_name)
parser.add_argument('-m', '--model', help='Model name to use as initializer.',
                    type=str, default='NONE')
parser.add_argument('-d', '--seed', help='Random seed used for generation.', 
                    type=int, default=42)
parser.add_argument('config', action='store', type=str)
# Compile and configure all the parameters
args = parser.parse_args()
# Set meta-parameters of the model
np.random.seed(args.seed)
# Logging configuration
# Set the basic configuration of the logging system
log_formatter = logging.Formatter(fmt='%(asctime)s [%(processName)s, %(process)s] [%(levelname)-5.5s]  %(message)s', 
                                  datefmt='%m-%d %H:%M')
logger = logging.getLogger()
logger.setLevel(logging.DEBUG)
# File logger
file_handler = logging.FileHandler("{}.log".format(args.name))
file_handler.setFormatter(log_formatter)
file_handler.setLevel(logging.DEBUG)
logger.addHandler(file_handler)
# Stderr logger
std_handler = logging.StreamHandler(sys.stdout)
std_handler.setFormatter(log_formatter)
std_handler.setLevel(logging.DEBUG)
logger.addHandler(std_handler)
# Training and test filename
# This script is used to train and test on the TREC dataset
# Since this data set is not explicitly partitioned into training/test 
# split, we use 10-fold cross-validation to report the final score
trec_train_filename = '../data/TREC/train_5500.label.txt'
trec_test_filename = '../data/TREC/TREC_10.label.txt'
# Label mapping
label_map = {'DESC':0, 'ENTY':1, 'ABBR':2, 'HUM':3, 'LOC':4, 'NUM':5}
# Loading and building training and test data set
trec_train_txt, trec_train_label = [], []
trec_test_txt, trec_test_label = [], []
start_time = time.time()
# Read all the instances 
with file(trec_train_filename, 'r') as fin:
    for line in fin:
        tokens = line.split()
        trec_train_txt.append(' '.join(tokens[1:]))
        trec_train_label.append(tokens[0].split(':')[0])
with file(trec_test_filename, 'r') as fin:
    for line in fin:
        tokens = line.split()
        trec_test_txt.append(' '.join(tokens[1:]))
        trec_test_label.append(tokens[0].split(':')[0])
# Shuffling all the data
assert len(trec_train_txt) == len(trec_train_label)
assert len(trec_test_txt) == len(trec_test_label)
train_size, test_size = len(trec_train_txt), len(trec_test_txt)
logger.debug('Size of the training data set: %d' % train_size)
logger.debug('Size of the test data set: %d' % test_size)
# Mapping string into int
trec_train_label = map(lambda x: label_map[x], trec_train_label)
trec_test_label = map(lambda x: label_map[x], trec_test_label)
# Random shuffling for training and test data set
rindex, tindex = np.arange(train_size), np.arange(test_size)
np.random.shuffle(rindex)
np.random.shuffle(tindex)
trec_train_txt = list(np.asarray(trec_train_txt)[rindex])
trec_train_label = list(np.asarray(trec_train_label)[rindex])
trec_test_txt = list(np.asarray(trec_test_txt)[tindex])
trec_test_label = list(np.asarray(trec_test_label)[tindex])
end_time = time.time()
logger.debug('Time used to load and shuffle the TREC training and test data set: %f seconds.' % (end_time-start_time))
# Load word-embedding
embedding_filename = '../data/wiki_embeddings.txt'
# Load training/test data sets and wiki-embeddings
word_embedding = WordEmbedding(embedding_filename)
embed_dim = word_embedding.embedding_dim()
start_time = time.time()
blank_index = word_embedding.word2index('</s>')
logger.debug('Blank index: {}'.format(word_embedding.index2word(blank_index)))
# Word-vector representation
trec_train_matrix, trec_train_sparse_select = [], []
trec_test_matrix, trec_test_sparse_select = [], []
# Record the sparse input indicator matrix only once for fast computation
for i, sent in enumerate(trec_train_txt):
    words = sent.split()
    words = [word.lower() for word in words]
    vectors = np.zeros((len(words)+2, embed_dim), dtype=np.float32)
    vectors[1:-1] = np.asarray([word_embedding.wordvec(word) for word in words])
    # Record the indices of words in each sentence
    indices = [blank_index]
    indices += [word_embedding.word2index(word) for word in words]
    indices += [blank_index]
    # Construct the sparse one-hot encoding matrix for fast updating word-embedding matrix
    sparse_select = lil_matrix((len(words)+2, word_embedding.dict_size()), dtype=floatX)
    sparse_select[range(len(words)+2), indices] = 1.0
    sparse_select = csc_matrix(sparse_select)
    # Append all the information
    trec_train_matrix.append(vectors)
    trec_train_sparse_select.append(sparse_select)
for i, sent in enumerate(trec_test_txt):
    words = sent.split()
    words = [word.lower() for word in words]
    vectors = np.zeros((len(words)+2, embed_dim), dtype=np.float32)
    vectors[1:-1] = np.asarray([word_embedding.wordvec(word) for word in words])
    # Record the indices of words in each sentence
    indices = [blank_index]
    indices += [word_embedding.word2index(word) for word in words]
    indices += [blank_index]
    # Construct the sparse one-hot encoding matrix for fast updating word-embedding matrix
    sparse_select = lil_matrix((len(words)+2, word_embedding.dict_size()), dtype=floatX)
    sparse_select[range(len(words)+2), indices] = 1.0
    sparse_select = csc_matrix(sparse_select)
    # Append all the information
    trec_test_matrix.append(vectors)
    trec_test_sparse_select.append(sparse_select)
end_time = time.time()
logger.debug('Time used to build sparse and dense input word-embedding matrices: %f seconds.' % (end_time-start_time))
# If there is a designated model, using it, else start from scratch
start_time = time.time()
configer = AdaSentConfiger(args.config)
if args.model == 'NONE':
    logger.debug('No designated model, training from scratch...')
    adasent = AdaSent(configer, verbose=True)
else:
    logger.debug('There is a designated model, loading: {}'.format(args.model))
    adasent = AdaSent.load(args.model)
end_time = time.time()
logger.debug('Time used to building the model: %f seconds.' % (end_time-start_time))
# Convert all the data into the form of numpy for easy access in the training set
trec_train_label, trec_train_matrix, trec_train_sparse_select = \
np.asarray(trec_train_label), np.asarray(trec_train_matrix), np.asarray(trec_train_sparse_select)
# Convert all the data into the form of numpy for easy access in the test set
trec_test_label, trec_test_matrix, trec_test_sparse_select = \
np.asarray(trec_test_label), np.asarray(trec_test_matrix), np.asarray(trec_test_sparse_select)
# Explicit partition of the training and test data set, so there is no need to 
# do cross-validation anymore
logger.debug('Start...')
# Initialize model training configuration
learn_rate = args.rate
batch_size = args.size
epoch = args.epoch
gradient_threshold = 10
embedding_gradient_threshold = 100
# Training using AdaGrad
training_threshold_epoch = epoch / 2
track_training_acc, track_training_cost = [], []
track_test_acc, track_test_cost = [], []
try:
    start_time = time.time()
    sample_size = 0
    fudge_factor = 1e-6
    # accumu matrix for word-embedding matrix
    # hist matrix for word-embedding matrix
    accumu_embedding = np.zeros((word_embedding.dict_size(), configer.num_input), dtype=floatX)
    hist_embedding = np.zeros((word_embedding.dict_size(), configer.num_input), dtype=floatX)
    # Training
    highest_train_accuracy, highest_test_accuracy = 0.0, 0.0
    for i in xrange(epoch):
        # # Random shuffle of the training samples at the beginning
        # rindex = np.random.permutation(train_size)
        # trec_train_label, trec_train_matrix, trec_train_sparse_select = \
        # trec_train_label[rindex], trec_train_matrix[rindex], trec_train_sparse_select[rindex]
        logger.debug('=' * 50)
        # Each epoch includes training on 9 folds and test on 1 fold
        costs = 0.0
        rate = learn_rate
        num_batch = train_size / batch_size
        correct_count = 0
        # Build 
        for k in xrange(num_batch):
            # Clear all the cache        
            accumu_grads = [np.zeros(param.get_value().shape, dtype=np.float32) for param in adasent.params]
            hist_grads = [np.zeros(param.get_value().shape, dtype=np.float32) for param in adasent.params]
            if i > training_threshold_epoch:
                accumu_embedding[:] = 0.0
                hist_embedding[:] = 0.0
            for j in xrange(k*batch_size, (k+1)*batch_size):
                train_sent_rep = trec_train_sparse_select[j].dot(word_embedding.embedding)
                results = adasent.compute_gradient_and_cost(train_sent_rep, trec_train_label[j])
                input_grad = adasent.compute_input_gradient(train_sent_rep, trec_train_label[j])
                grads, cost, pred = results[:-2], results[-2], results[-1]
                if pred == trec_train_label[j]: correct_count += 1
                costs += cost
                for accumu_grad, hist_grad, grad in zip(accumu_grads, hist_grads, grads):
                    accumu_grad += grad
                    hist_grad += np.square(grad)
                ## Update the word-embedding matrix
                if i > training_threshold_epoch:
                    tmp = trec_train_sparse_select[j].T.dot(input_grad)
                    accumu_embedding += tmp
                    hist_embedding += np.square(tmp)
            # Normalizing model parameters
            for accumu_grad, hist_grad in zip(accumu_grads, hist_grads):
                accumu_grad /= train_size-num_batch*batch_size
                accumu_grad /= fudge_factor + np.sqrt(hist_grad)
                # Gradient clipping
                accumu_norm = np.sqrt(np.sum(accumu_grad ** 2))
                if accumu_norm > gradient_threshold:
                    accumu_grad /= accumu_norm / gradient_threshold
            # Normalizing and updating word-embedding matrix
            if i > training_threshold_epoch:
                accumu_embedding /= train_size-num_batch*batch_size
                accumu_embedding /= fudge_factor + np.sqrt(hist_embedding)
                # Gradient clipping
                accumu_embedding_norm = np.sqrt(np.sum(accumu_embedding ** 2))
                if accumu_embedding_norm > embedding_gradient_threshold:
                    accumu_embedding /= accumu_embedding_norm / embedding_gradient_threshold
                word_embedding._embedding -= rate * accumu_embedding
            # Updating all the parameters
            adasent.update_params(accumu_grads, rate)
        # Clear all the cache again
        accumu_grads = [np.zeros(param.get_value().shape, dtype=np.float32) for param in adasent.params]
        hist_grads = [np.zeros(param.get_value().shape, dtype=np.float32) for param in adasent.params]
        if i > training_threshold_epoch:
            accumu_embedding[:] = 0.0
            hist_embedding[:] = 0.0
        if num_batch * batch_size < train_size:
            for j in xrange(num_batch*batch_size, train_size):
                train_sent_rep = trec_train_sparse_select[j].dot(word_embedding.embedding)
                results = adasent.compute_gradient_and_cost(train_sent_rep, trec_train_label[j])
                input_grad = adasent.compute_input_gradient(train_sent_rep, trec_train_label[j])
                grads, cost, pred = results[:-2], results[-2], results[-1]
                if pred == trec_train_label[j]: correct_count += 1
                costs += cost
                for accumu_grad, hist_grad, grad in zip(accumu_grads, hist_grads, grads):
                    accumu_grad += grad
                    hist_grad += np.square(grad)
                ## Update the word-embedding matrix
                if i > training_threshold_epoch:
                    tmp = trec_train_sparse_select[j].T.dot(input_grad)
                    accumu_embedding += tmp
                    hist_embedding += np.square(tmp)
            # Normalizing model parameters
            for accumu_grad, hist_grad in zip(accumu_grads, hist_grads):
                accumu_grad /= train_size-num_batch*batch_size
                accumu_grad /= fudge_factor + np.sqrt(hist_grad)
                # Gradient clipping
                accumu_norm = np.sqrt(np.sum(accumu_grad ** 2))
                if accumu_norm > gradient_threshold:
                    accumu_grad /= accumu_norm / gradient_threshold
            # Normalizing and updating word-embedding matrix
            if i > training_threshold_epoch:
                accumu_embedding /= train_size-num_batch*batch_size
                accumu_embedding /= fudge_factor + np.sqrt(hist_embedding)
                # Gradient clipping
                accumu_embedding_norm = np.sqrt(np.sum(accumu_embedding ** 2))
                if accumu_embedding_norm > embedding_gradient_threshold:
                    accumu_embedding /= accumu_embedding_norm / embedding_gradient_threshold
                word_embedding._embedding -= rate * accumu_embedding
            # Updating all the parameters
            adasent.update_params(accumu_grads, rate)
        train_accuracy = float(correct_count) / train_size
        logger.debug('Training epoch: %d, total cost: %f, accuracy = %f' 
                    % (i, costs, train_accuracy))
        # Append all the numbers
        track_training_cost.append(costs)
        track_training_acc.append(train_accuracy)
        if train_accuracy > highest_train_accuracy: highest_train_accuracy = train_accuracy
        # Testing
        correct_count = 0
        costs = 0.0
        for j in xrange(test_size):
            test_sent_rep = trec_test_sparse_select[j].dot(word_embedding.embedding)
            pred = adasent.predict(test_sent_rep)
            cost = adasent.show_cost(test_sent_rep, trec_test_label[j])
            if pred == trec_test_label[j]: correct_count += 1
            costs += cost
        test_accuracy = float(correct_count) / test_size
        logger.debug('Test accuracy: %f' % test_accuracy)
        # Append all the numbers
        track_test_cost.append(costs)
        track_test_acc.append(test_accuracy)
        if test_accuracy > highest_test_accuracy: highest_test_accuracy = test_accuracy
        # Check norms of the model parameter
        for param in adasent.params:
            val = param.get_value(borrow=True)
            norm = np.sqrt(np.sum(np.square(val)))
            logger.debug('Parameter: {}, L2-norm: {}'.format(param.name, norm))
        wnorm = np.sqrt(np.sum(np.square(word_embedding._embedding)))
        logger.debug('Parameter: {}, L2-norm: {}'.format('Word-Embedding', wnorm))    
except:
    logger.debug('Error appeared!')
    traceback.print_exc(file=sys.stdout)
    logger.debug('-' * 50)
finally:
    end_time = time.time()
    logger.debug('Time used for training: %f seconds.' % (end_time-start_time))
    AdaSent.save('{}-adasent.model'.format(args.name), adasent)
    # Output best result
    logger.debug('Highest Training Accuracy: %f' % highest_train_accuracy)
    logger.debug('Highest Test Accuracy: %f' % highest_test_accuracy)
    logger.debug('Hyperparameter used for this experiment: {}'.format(sys.argv))
    # Save all the tracking numbers
    logger.debug('Training and test finished, model saved at: {}-adasent.model'.format(args.name))
    logger.debug('Finished...')
