#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Date    : 2015-01-02 23:50:58
# @Author  : Han Zhao (han.zhao@uwaterloo.ca)
# @Link    : https://github.com/KeiraZhao
# @Version : $Id$

import os, sys
import numpy as np
import scipy as sp
import theano.tensor as T
import theano
import time 
import cPickle
import logging

import config
import utils
from config import NBoWConfiger
from utils import floatX
from activations import Activation
from logistic import SoftmaxLayer, LogisticLayer
from mlp import HiddenLayer

logger = logging.getLogger(__name__)


class NBoW(object):
    '''
    Neural Bag-of-Words Model for phrases/sentences/short texts classification.
    Apply global pooling to obtain the input vector, either average pooling or max pooling.
    Then followed by typical MLP architecture.
    '''
    def __init__(self, config, verbose=True):
        if verbose: logger.debug('Building NBoW Classifier...')
        self.input = T.matrix(name='NBoW_input')
        # Configure Activation function
        self.act = Activation(config.hiddenact)
        self.pooling = config.pooling
        # Compute the number of parameters in this model
        self.num_params = config.num_input * config.num_mlp + config.num_mlp + \
                          config.num_mlp * config.num_class + config.num_class
        # Initialize model parameters
        np.random.seed(config.random_seed)
        # Global pooling
        if config.pooling == 'averaging':
            self.pooled_input = T.mean(self.input, axis=0)
        elif config.pooling == 'max':
            self.pooled_input = T.max(self.input, axis=0)
        else:
            raise NotImplementedError('Not implemented pooling method: {}'.format(config.pooling))
        # MLP component
        self.hidden_layer = HiddenLayer(self.pooled_input, (config.num_input, config.num_mlp),
                                        act=self.act)
        self.hidden = self.hidden_layer.output
        # Dropout parameter
        srng = T.shared_randomstreams.RandomStreams(config.random_seed)
        mask = srng.binomial(n=1, p=1-config.dropout, size=self.hidden.shape)
        self.hidden *= T.cast(mask, floatX)
        # Softmax Classifier
        self.output_layer = SoftmaxLayer(self.hidden, (config.num_mlp, config.num_class))
        # Since we'd like to use AdaGrad to fine-tuning the word-embedding matrix, we need to 
        # compute in online version
        self.output = self.output_layer.output[0]
        self.pred = self.output_layer.pred[0]
        # Stack parameters
        self.params = []
        self.params += self.hidden_layer.params
        self.params += self.output_layer.params
        # Build objective functions
        self.truth = T.iscalar(name='label')
        self.cost = -T.log(self.output)[self.truth]
        ## Weight Decay
        if config.weight_decay:
            self.regularizer = self.hidden_layer.L2_loss() + self.output_layer.L2_loss()
            self.regularizer *= config.weight_decay_parameter
            self.cost += self.regularizer
        # Construct gradient vector
        self.gradparams = T.grad(self.cost, self.params)
        # Construct gradient for the input matrix, for fine-tuning
        self.input_grads = T.grad(self.cost, self.input)
        # Build and compile theano functions
        self.predict = theano.function(inputs=[self.input], outputs=self.pred)
        self.objective = theano.function(inputs=[self.input, self.truth], outputs=self.cost)
        self.compute_gradient_and_cost = theano.function(inputs=[self.input, self.truth],
                    outputs=self.gradparams+[self.cost, self.pred])
        self.compute_input_gradient = theano.function(inputs=[self.input, self.truth], 
                    outputs=self.input_grads)
        # Theano function for debugging purpose
        self.show_hidden = theano.function(inputs=[self.input], outputs=self.hidden)
        self.show_prob = theano.function(inputs=[self.input], outputs=self.output)
        self.show_cost = theano.function(inputs=[self.input, self.truth], outputs=self.cost)
        if verbose:
            logger.debug('{} built finished...'.format(self.__class__.__name__))
            logger.debug('Global pooling function used: {}'.format(config.pooling))
            logger.debug('Hidden activation function used: {}'.format(config.hiddenact))
            logger.debug('Total number of parameters in the model: {}'.format(self.num_params))

    def update_params(self, grads, learn_rate):
        '''
        @grads: [np.ndarray]. List of numpy.ndarray to update the model parameters.
        @learn_rate: scalar. Learning rate.
        '''
        for param, grad in zip(self.params, grads):
            p = param.get_value(borrow=True)
            param.set_value(p - learn_rate * grad, borrow=True)

    def set_params(self, params):
        '''
        @params: [np.ndarray]. List of numpy.ndarray to set the model parameters.
        '''
        for p, param in zip(self.params, params):
            p.set_value(param, borrow=True)

    def deepcopy(self, nbow):
        '''
        @nbow: An instance of NBoW. Copy the model parameters.
        '''
        assert len(self.params) == len(nbow.params)

    @staticmethod
    def save(fname, model):
        '''
        @fname: String. File name to store the model.
        @model: An instance of NBoW to be saved.
        '''
        with file(fname, 'wb') as fout:
            cPickle.dump(model, fout)

    @staticmethod
    def load(fname):
        '''
        @fname: String. File name to load the model.
        '''
        with file(fname, 'rb') as fin:
            model = cPickle.load(fin)
        return model

