#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Date    : 2014-08-29 10:37:00
# @Author  : Han Zhao (han.zhao@uwaterloo.ca)
# @Link    : https://github.com/KeiraZhao
# @Version : 0.0
import os, sys
import numpy as np
import theano
import theano.tensor as T
import time
import cPickle
import logging
from pprint import pprint

import config
import utils
from utils import floatX
from activations import Activation
from logistic import SoftmaxLayer, LogisticLayer
from score import ScoreLayer
from mlp import HiddenLayer


logger = logging.getLogger(__name__)

class RNNEncoder(object):
	'''
	Single directional RNN for sequence encoding. In order to collapse a sequence of hidden 
	vector representations into a fixed-length hidden representation, either T.mean or T.max is 
	used as pooling operation.
	'''
	def __init__(self, config, verbose=True):
		if verbose: logger.debug('Building Vanilla RNN Encoder...')
		self.input = T.matrix(name='RNNEncoder_input')
		# Configure Activation function
		self.act = Activation(config.activation)
		# Build Vanilla RNN
		num_input, num_hidden = config.num_input, config.num_hidden
		self.num_params = num_input * num_hidden + num_hidden * num_hidden + num_hidden
		# Initialize model parameters
		np.random.seed(config.random_seed)
		# Feed-forward matrix
		W_val = np.random.uniform(low=-1.0, high=1.0, size=(num_input, num_hidden)).astype(floatX)
		self.W = theano.shared(value=W_val, name='W', borrow=True)
		# Recurrent matrix
		U_val = np.random.uniform(low=-1.0, high=1.0, size=(num_hidden, num_hidden)).astype(floatX)
		self.U = theano.shared(value=U_val, name='U', borrow=True)
		# Bias parameter
		b_val = np.zeros(num_hidden, dtype=floatX)
		self.b = theano.shared(value=b_val, name='b', borrow=True)
		# h0 vectors, treated as constants
		self.h0 = theano.shared(value=np.zeros(num_hidden, dtype=floatX), name='h0', borrow=True)
		# Stack all the parameters
		self.params = [self.W, self.U, self.b]
		# Compute the hidden chain representation over time
		self.hs, _ = theano.scan(fn=self._step, 
								 sequences=self.input,
								 outputs_info=[self.h0],
								 truncate_gradient=config.bptt)
		# Use pooling technique along time to reduce into a fixed-length representation
		# Configure based on the option
		self.pooling = config.pooling
		if self.pooling == 'max':
			self.output = T.max(self.hs, axis=0)
		elif self.pooling == 'averaging':
			self.output = T.mean(self.hs, axis=0)
		elif self.pooling == 'final':
			self.output = self.hs[-1]
		else:
			raise NotImplementedError('Not implemented pooling method: {}'.format(self.pooling))
		# L1, L2 regularization
		self.L1_norm = T.sum(T.abs_(self.W)) + T.sum(T.abs_(self.U))
		self.L2_norm = T.sum(self.W ** 2) + T.sum(self.U ** 2)
		if verbose:
			logger.debug('Finished constructing the structure of {}: '.format(self.__class__.__name__))
			logger.debug('Size of the input dimension: %d' % config.num_input)
			logger.debug('Size of the hidden dimension: %d' % config.num_hidden)
			logger.debug('Activation function: %s' % config.activation)
			logger.debug('Pooling method: %s' % config.pooling)
			logger.debug('Number of parameters used: %d' % self.num_params)

	def _step(self, x_t, h_tm1):
		h_t = self.act.activate(T.dot(x_t, self.W) + \
								T.dot(h_tm1, self.U) + \
								self.b)
		return h_t

	def L1_loss(self):
		return self.L1_norm

	def L2_loss(self):
		'''
		Return L2 norm of the model parameters
		'''
		return self.L2_norm

	def encode(self, inputM):
		'''
		@inputM: Theano symbolic matrix. Compress the input matrix into output vector.
		'''
		hs, _ = theano.scan(fn=self._step, 
							sequences=inputM,
							outputs_info=[self.h0])
		# Dynamic pooling
		if self.pooling == 'max':
			return T.max(hs, axis=0)
		elif self.pooling == 'averaging':
			return T.mean(hs, axis=0)
		else:
			raise NotImplementedError('Not implemented pooling method: {}'.format(self.pooling))


class BRNNEncoder(object):
	'''
	Bidirectional RNN for sequence encoding. In order to collapse a sequence of hidden 
	vector representations into a fixed-length hidden representation, either T.mean or T.max is 
	used as pooling operation.
	'''
	def __init__(self, config, verbose=True):
		if verbose: logger.debug('Building Bidirectional RNN Encoder...')
		self.input = T.matrix(name='BRNNEncoder_input')
		# Configure Activation function
		self.act = Activation(config.activation)
		# Build Bidirectional RNN
		num_input, num_hidden = config.num_input, config.num_hidden
		self.num_params = 2 * (num_input * num_hidden + num_hidden * num_hidden + num_hidden)
		# Initialize model parameters
		np.random.seed(config.random_seed)
		# 1, Feed-forward matrix for forward direction: W_forward
		W_forward_val = np.random.uniform(low=-1.0, high=1.0, size=(num_input, num_hidden))
		W_forward_val = W_forward_val.astype(floatX)
		self.W_forward = theano.shared(value=W_forward_val, name='W_forward', borrow=True)
		# 1, Feed-forward matrix for backward direction: W_backward
		W_backward_val = np.random.uniform(low=-1.0, high=1.0, size=(num_input, num_hidden))
		W_backward_val = W_backward_val.astype(floatX)
		self.W_backward = theano.shared(value=W_backward_val, name='W_backward', borrow=True)
		# 2, Recurrent matrix for forward direction: U_forward
		U_forward_val = np.random.uniform(low=-1.0, high=1.0, size=(num_hidden, num_hidden))
		U_forward_val = U_forward_val.astype(floatX)
		U_forward_val, _, _ = np.linalg.svd(U_forward_val)
		self.U_forward = theano.shared(value=U_forward_val, name='U_forward', borrow=True)
		# 2, Recurrent matrix for backward direction: U_backward
		U_backward_val = np.random.uniform(low=-1.0, high=1.0, size=(num_hidden, num_hidden))
		U_backward_val = U_backward_val.astype(floatX)
		U_backward_val, _, _ = np.linalg.svd(U_backward_val)
		self.U_backward = theano.shared(value=U_backward_val, name='U_backward', borrow=True)
		# 3, Bias parameter for the hidden-layer forward direction RNN
		b_forward_val = np.zeros(num_hidden, dtype=floatX)
		self.b_forward = theano.shared(value=b_forward_val, name='b_forward', borrow=True)
		# 3, Bias parameter for the hidden-layer backward direction RNN
		b_backward_val = np.zeros(num_hidden, dtype=floatX)
		self.b_backward = theano.shared(value=b_backward_val, name='b_backward', borrow=True)
		# h[0], zero vectors, treated as constants
		self.h0_forward = theano.shared(value=np.zeros(num_hidden, dtype=floatX), name='h0_forward', borrow=True)
		self.h0_backward = theano.shared(value=np.zeros(num_hidden, dtype=floatX), name='h0_backward', borrow=True)
		# Stack all the parameters
		self.params = [self.W_forward, self.W_backward, self.U_forward, self.U_backward, 
					   self.b_forward, self.b_backward]
		# Compute the forward and backward representation over time
		self.h_forwards, _ = theano.scan(fn=self._forward_step, 
										 sequences=self.input, 
										 outputs_info=[self.h0_forward],
										 truncate_gradient=config.bptt)
		self.h_backwards, _ = theano.scan(fn=self._backward_step,
										  sequences=self.input,
										  outputs_info=[self.h0_backward],
										  truncate_gradient=config.bptt,
										  go_backwards=True)
		# Dynamic Pooling based on the choice of input configuration file
		self.pooling = config.pooling
		if self.pooling == 'max':
			self.h_forward = T.max(self.h_forwards, axis=0)
			self.h_backward = T.max(self.h_backwards, axis=0)
		elif self.pooling == 'averaging':
			self.h_forward = T.mean(self.h_forwards, axis=0)
			self.h_backward = T.mean(self.h_backwards, axis=0)
		elif self.pooling == 'final':
			self.h_forward = self.h_forwards[-1]
			self.h_backward = self.h_backwards[-1]
		else:
			raise NotImplementedError('Not implemented pooling method: {}'.format(self.pooling))
		# Concatenate
		self.output = T.concatenate([self.h_forward, self.h_backward], axis=0)
		# L1, L2 regularization
		self.L1_norm = T.sum(T.abs_(self.W_forward)) + T.sum(T.abs_(self.W_backward)) + \
					   T.sum(T.abs_(self.U_forward)) + T.sum(T.abs_(self.U_backward))
		self.L2_norm = T.sum(self.W_forward ** 2) + T.sum(self.W_backward ** 2) + \
					   T.sum(self.U_forward ** 2) + T.sum(self.U_backward ** 2)
		if verbose:
			logger.debug('Finished constructing the structure of {}: '.format(self.__class__.__name__))
			logger.debug('Size of the input dimension: %d' % num_input)
			logger.debug('Size of the hidden dimension: %d' % num_hidden)
			logger.debug('Activation function: %s' % config.activation)
			logger.debug('Pooling method: %s' % config.pooling)
			logger.debug('Number of parameters used: %d' % self.num_params)

	def _forward_step(self, x_t, h_tm1):
		h_t = self.act.activate(T.dot(x_t, self.W_forward) + \
								T.dot(h_tm1, self.U_forward) + \
								self.b_forward)
		return h_t

	def _backward_step(self, x_t, h_tm1):
		h_t = self.act.activate(T.dot(x_t, self.W_backward) + \
								T.dot(h_tm1, self.U_backward) + \
								self.b_backward)
		return h_t

	def L1_loss(self):
		return self.L1_norm

	def L2_loss(self):
		return self.L2_norm

	def encode(self, inputM):
		'''
		@inputM: Theano symbol matrix. Compress the input matrix into output vector.
		'''
		h_forwards, _ = theano.scan(fn=self._forward_step, 
									sequences=inputM,
									outputs_info=[self.h0_forward])
		h_backwards, _ = theano.scan(fn=self._backward_step, 
									 sequences=inputM,
									 outputs_info=[self.h0_backward],
									 go_backwards=True)
		# Dynamic pooling
		if self.pooling == 'max':
			h_forward = T.max(h_forwards, axis=0)
			h_backward = T.max(h_backwards, axis=0)
		elif self.pooling == 'averaging':
			h_forward = T.mean(h_forwards, axis=0)
			h_backward = T.mean(h_backwards, axis=0)
		else:
			raise NotImplementedError('Not implemented pooling method: {}'.format(self.pooling))
		# Concatenate
		h = T.concatenate([h_forward, h_backward], axis=0)
		return h


class RNNClassifier(object):
	'''
	Vanilla RNN for time-series classification problem.
	'''
	def __init__(self, config, verbose=True):
		# Construct an RNN Encoder
		self.encoder = self.build_encoder(config, verbose)
		self.input = self.encoder.input
		# Get hidden representation from the RNN based sequence encoder
		self.hidden = self.encoder.output
		# Stack parameters
		self.params = []
		self.params += self.encoder.params
		# Construct activation functions
		self.act = Activation(config.activation)
		# MLP Component
		self.hidden_layer = HiddenLayer(self.hidden, (config.num_hidden, config.num_mlp), 
										act=Activation(config.hiddenact))
		self.compressed_hidden = self.hidden_layer.output
		self.params += self.hidden_layer.params
		# Dropout parameter
		srng = T.shared_randomstreams.RandomStreams(config.random_seed)
		mask = srng.binomial(n=1, p=1-config.dropout, size=self.compressed_hidden.shape)
		self.compressed_hidden *= T.cast(mask, floatX)
		# Softmax classifier
		self.output_layer = SoftmaxLayer(self.compressed_hidden, (config.num_mlp, config.num_class))
		self.output = self.output_layer.output[0]
		self.params += self.output_layer.params
		self.pred = self.output_layer.pred[0]
		# Compute the total number of parameters in the model
		self.num_params = self.encoder.num_params + self.hidden_layer.num_params + self.output_layer.num_params
		# Build objective function
		self.truth = T.iscalar(name='label')
		self.cost = -T.log(self.output)[self.truth]
		## Weight Decay
		if config.weight_decay:
			self.regularizer = self.encoder.L2_loss() + self.hidden_layer.L2_loss() + \
							   self.output_layer.L2_loss()
			self.regularizer *= config.weight_decay_parameter
			self.cost += self.regularizer
		# Construct gradient vectors
		self.gradparams = T.grad(self.cost, self.params)
		# Construct gradient for the input matrix, for fine-tuning
		self.input_grads = T.grad(self.cost, self.input)
		# Build and compile theano functions
		self.predict = theano.function(inputs=[self.input], outputs=self.pred)
		self.compute_gradient_and_cost = theano.function(inputs=[self.input, self.truth], 
					outputs=self.gradparams+[self.cost, self.pred])
		self.compute_input_gradient = theano.function(inputs=[self.input, self.truth], 
					outputs=self.input_grads)
		# Theano function for debugging purpose
		self.show_hidden = theano.function(inputs=[self.input], outputs=self.hidden)
		self.show_compressed_hidden = theano.function(inputs=[self.input], outputs=self.compressed_hidden)
		self.show_prob = theano.function(inputs=[self.input], outputs=self.output)
		self.show_cost = theano.function(inputs=[self.input, self.truth], outputs=self.cost)
		if verbose:
			logger.debug('{} built finished...'.format(self.__class__.__name__))
			logger.debug('Total number of parameters in the model: %d' % self.num_params)

	def build_encoder(self, config, verbose):
		'''
		@config: RNNConfiger. This method should be overwritten by subclasses to reuse.
		'''
		return RNNEncoder(config, verbose)

	def update_params(self, grads, learn_rate):
		'''
		@grads: [np.ndarray]. List of numpy.ndarray to update the model parameters.
		@learn_rate: scalar. Learning rate.
		'''
		for param, grad in zip(self.params, grads):
			p = param.get_value(borrow=True)
			param.set_value(p - learn_rate * grad, borrow=True)

	def set_parmas(self, parmas):
		'''
		@params: [np.ndarray]. List of numpy.ndarray to set the model parameters.
		'''
		for p, param in zip(self.params, params):
			p.set_value(param, borrow=True)

	def deepcopy(self, rnnclassifier):
		'''
		@rnnclassifier: An instance of RNNClassifier. Copy the model parameters of another RNNClassifier and use it.
		'''
		assert len(self.params) == len(rnnclassifier.params)
		for p, param in zip(self.params, rnnclassifier.params):
			val = param.get_value()
			p.set_value(val, borrow=True)

	@staticmethod
	def save(fname, model):
		'''
		@fname: String. File name to store the model.
		@model: An instance of RNNClassifier to be saved.
		'''
		with file(fname, 'wb') as fout:
			cPickle.dump(model, fout)

	@staticmethod
	def load(fname):
		'''
		@fname: String. File name to load the model.
		'''
		with file(fname, 'rb') as fin:
			model = cPickle.load(fin)
		return model


class BRNNClassifier(RNNClassifier):
	def __init__(self, config, verbose=True):
		# Construct an BRNN Encoder
		self.encoder = self.build_encoder(config, verbose)
		self.input = self.encoder.input
		# Get hidden representation from the RNN based sequence encoder
		self.hidden = self.encoder.output
		# Stack parameters
		self.params = []
		self.params += self.encoder.params
		# Construct activation functions
		self.act = Activation(config.activation)
		# MLP Component
		self.hidden_layer = HiddenLayer(self.hidden, (2 * config.num_hidden, config.num_mlp), 
										act=Activation(config.hiddenact))
		self.compressed_hidden = self.hidden_layer.output
		self.params += self.hidden_layer.params
		# Dropout parameter
		srng = T.shared_randomstreams.RandomStreams(config.random_seed)
		mask = srng.binomial(n=1, p=1-config.dropout, size=self.compressed_hidden.shape)
		self.compressed_hidden *= T.cast(mask, floatX)
		# Softmax classifier 
		self.output_layer = SoftmaxLayer(self.compressed_hidden, (config.num_mlp, config.num_class))
		self.output = self.output_layer.output[0]
		self.params += self.output_layer.params
		self.pred = self.output_layer.pred[0]
		# Compute the total number of parameters in the model
		self.num_params = self.encoder.num_params + self.hidden_layer.num_params + self.output_layer.num_params
		# Build objective function
		self.truth = T.iscalar(name='label')
		self.cost = -T.log(self.output)[self.truth]
		## Weight Decay
		if config.weight_decay:
			self.regularizer = self.encoder.L2_loss() + self.hidden_layer.L2_loss() + \
							   self.output_layer.L2_loss()
			self.regularizer *= config.weight_decay_parameter
			self.cost += self.regularizer
		# Construct gradient vectors
		self.gradparams = T.grad(self.cost, self.params)
		# Construct gradient for the input matrix, for fine-tuning
		self.input_grads = T.grad(self.cost, self.input)
		# Build and compile theano functions
		self.predict = theano.function(inputs=[self.input], outputs=self.pred)
		self.compute_gradient_and_cost = theano.function(inputs=[self.input, self.truth], 
					outputs=self.gradparams+[self.cost, self.pred])
		self.compute_input_gradient = theano.function(inputs=[self.input, self.truth], 
					outputs=self.input_grads)
		# Theano function for debugging purpose
		self.show_hidden = theano.function(inputs=[self.input], outputs=self.hidden)
		self.show_compressed_hidden = theano.function(inputs=[self.input], outputs=self.compressed_hidden)
		self.show_prob = theano.function(inputs=[self.input], outputs=self.output)
		self.show_cost = theano.function(inputs=[self.input, self.truth], outputs=self.cost)
		if verbose:
			logger.debug('{} built finished...'.format(self.__class__.__name__))
			logger.debug('Total number of parameters in the model: %d' % self.num_params)

	def build_encoder(self, config, verbose):
		'''
		@config: RNNConfiger. 
		'''
		return BRNNEncoder(config, verbose)


class BRNNMatcher(object):
	'''
	Bidirectional RNN for text matching as a classification problem.
	'''
	def __init__(self, config, verbose=True):
		# Construct two BRNNEncoders for matching two sentences
		self.encoderL = BRNNEncoder(config, verbose)
		self.encoderR = BRNNEncoder(config, verbose)
		# Link two parts
		self.params = []
		self.params += self.encoderL.params
		self.params += self.encoderR.params
		# Set up input
		self.inputL = self.encoderL.input
		self.inputR = self.encoderR.input
		# Get output of two BRNNEncoders
		self.hiddenL = self.encoderL.output
		self.hiddenR = self.encoderR.output
		# Activation function
		self.act = Activation(config.activation)
		# MLP Component
		self.hidden = T.concatenate([self.hiddenL, self.hiddenR], axis=0)
		self.hidden_layer = HiddenLayer(self.hidden, 
										(4*config.num_hidden, config.num_mlp), 
										act=Activation(config.hiddenact))
		self.compressed_hidden = self.hidden_layer.output
		# Accumulate parameters
		self.params += self.hidden_layer.params
		# Dropout parameter
		srng = T.shared_randomstreams.RandomStreams(config.random_seed)
		mask = srng.binomial(n=1, p=1-config.dropout, size=self.compressed_hidden.shape)
		self.compressed_hidden *= T.cast(mask, floatX)
		# Logistic regression
		self.logistic_layer = LogisticLayer(self.compressed_hidden, config.num_mlp)
		self.output = self.logistic_layer.output
		self.pred = self.logistic_layer.pred
		# Accumulate parameters
		self.params += self.logistic_layer.params
		# Compute the total number of parameters in the model
		self.num_params_encoder = self.encoderL.num_params + self.encoderR.num_params
		self.num_params_classifier = 2 * config.num_hidden * config.num_mlp + config.num_mlp + \
									 config.num_mlp + 1
		self.num_params = self.num_params_encoder + self.num_params_classifier
		# Build target function
		self.truth = T.ivector(name='label')
		self.cost = self.logistic_layer.NLL_loss(self.truth)
		# Build computational graph and compute the gradients of the model parameters
		# with respect to the cost function
		self.gradparams = T.grad(self.cost, self.params)
		# Compile theano function
		self.objective = theano.function(inputs=[self.inputL, self.inputR, self.truth], outputs=self.cost)
		self.predict = theano.function(inputs=[self.inputL, self.inputR], outputs=self.pred)
		# Compute the gradient of the objective function and cost and prediction
		self.compute_cost_and_gradient = theano.function(inputs=[self.inputL, self.inputR, self.truth],
														 outputs=self.gradparams+[self.cost, self.pred])
		# Output function for debugging purpose
		self.show_hidden = theano.function(inputs=[self.inputL, self.inputR], outputs=self.hidden)
		self.show_compressed_hidden = theano.function(inputs=[self.inputL, self.inputR], outputs=self.compressed_hidden)
		self.show_output = theano.function(inputs=[self.inputL, self.inputR], outputs=self.output)
		if verbose:
			logger.debug('Architecture of {} built finished, summarized below: '.format(self.__class__.__name__))
			logger.debug('Input dimension: %d' % config.num_input)
			logger.debug('Hidden dimension of RNN: %d' % config.num_hidden)
			logger.debug('Hidden dimension of MLP: %d' % config.num_mlp)
			logger.debug('Number of parameters in the encoder part: %d' % self.num_params_encoder)
			logger.debug('Number of parameters in the classifier: %d' % self.num_params_classifier)
			logger.debug('Total number of parameters in this model: %d' % self.num_params)

	def update_params(self, grads, learn_rate):
		'''
		@grads: [np.ndarray]. List of numpy.ndarray for updating the model parameters.
				They are the corresponding gradients of model parameters.
		@learn_rate: scalar. Learning rate.
		'''
		for param, grad in zip(self.params, grads):
			p = param.get_value(borrow=True)
			param.set_value(p - learn_rate * grad, borrow=True)

	@staticmethod
	def save(fname, model):
		'''
		@fname: String. Filename to store the model.
		@model: BRNNMatcher. An instance of BRNNMatcher to be saved.
		'''
		with file(fname, 'wb') as fout:
			cPickle.dump(model, fout)

	@staticmethod
	def load(fname):
		'''
		@fname: String. Filename to load the model.
		'''
		with file(fname, 'rb') as fin:
			model = cPickle.load(fin)
		return model


class BRNNMatchScorer(object):
	'''
	Bidirectional RNN for text matching as a classification problem.
	'''
	def __init__(self, config, verbose=True):
		# Construct two BRNNEncoders for matching two sentences
		self.encoderL = BRNNEncoder(config, verbose)
		self.encoderR = BRNNEncoder(config, verbose)
		# Link two parts
		self.params = []
		self.params += self.encoderL.params
		self.params += self.encoderR.params
		# Set up input
		# Note that there are three kinds of inputs altogether, including:
		# 1, inputL, inputR. This pair is used for computing the score after training
		# 2, inputPL, inputPR. This pair is used for training positive pairs
		# 3, inputNL, inputNR. This pair is used for training negative pairs
		self.inputL = self.encoderL.input
		self.inputR = self.encoderR.input
		# Positive 
		self.inputPL = T.matrix(name='inputPL', dtype=floatX)
		self.inputPR = T.matrix(name='inputPR', dtype=floatX)
		# Negative
		self.inputNL = T.matrix(name='inputNL', dtype=floatX)
		self.inputNR = T.matrix(name='inputNR', dtype=floatX)
		# Get output of two BRNNEncoders
		self.hiddenL = self.encoderL.output
		self.hiddenR = self.encoderR.output
		# Positive Hidden
		self.hiddenPL = self.encoderL.encode(self.inputPL)
		self.hiddenPR = self.encoderR.encode(self.inputPR)
		# Negative Hidden
		self.hiddenNL = self.encoderL.encode(self.inputNL)
		self.hiddenNR = self.encoderR.encode(self.inputNR)
		# Activation function
		self.act = Activation(config.activation)
		self.hidden = T.concatenate([self.hiddenL, self.hiddenR], axis=0)
		self.hiddenP = T.concatenate([self.hiddenPL, self.hiddenPR], axis=0)
		self.hiddenN = T.concatenate([self.hiddenNL, self.hiddenNR], axis=0)
		# Build hidden layer
		self.hidden_layer = HiddenLayer(self.hidden, 
										(4*config.num_hidden, config.num_mlp), 
										act=Activation(config.hiddenact))
		self.compressed_hidden = self.hidden_layer.output
		self.compressed_hiddenP = self.hidden_layer.encode(self.hiddenP)
		self.compressed_hiddenN = self.hidden_layer.encode(self.hiddenN)
		# Accumulate parameters
		self.params += self.hidden_layer.params
		# Dropout parameter
		srng = T.shared_randomstreams.RandomStreams(config.random_seed)
		mask = srng.binomial(n=1, p=1-config.dropout, size=self.compressed_hidden.shape)
		maskP = srng.binomial(n=1, p=1-config.dropout, size=self.compressed_hiddenP.shape)
		maskN = srng.binomial(n=1, p=1-config.dropout, size=self.compressed_hiddenN.shape)
		self.compressed_hidden *= T.cast(mask, floatX)
		self.compressed_hiddenP *= T.cast(maskP, floatX)
		self.compressed_hiddenN *= T.cast(maskN, floatX)
		# Score layer
		self.score_layer = ScoreLayer(self.compressed_hidden, config.num_mlp)
		self.output = self.score_layer.output
		self.scoreP = self.score_layer.encode(self.compressed_hiddenP)
		self.scoreN = self.score_layer.encode(self.compressed_hiddenN)
		# Accumulate parameters
		self.params += self.score_layer.params
		# Build cost function
		self.cost = T.mean(T.maximum(T.zeros_like(self.scoreP), 1.0 - self.scoreP + self.scoreN))
		# Construct the total number of parameters in the model
		self.gradparams = T.grad(self.cost, self.params)
		# Compute the total number of parameters in the model
		self.num_params_encoder = self.encoderL.num_params + self.encoderR.num_params
		self.num_params_classifier = 2 * config.num_hidden * config.num_mlp + config.num_mlp + \
									 config.num_mlp + 1
		self.num_params = self.num_params_encoder + self.num_params_classifier
		# Build class functions
		self.score = theano.function(inputs=[self.inputL, self.inputR], outputs=self.output)
		# Compute the gradient of the objective function and cost and prediction
		self.compute_cost_and_gradient = theano.function(inputs=[self.inputPL, self.inputPR, 
																 self.inputNL, self.inputNR],
														 outputs=self.gradparams+[self.cost, self.scoreP, self.scoreN])
		# Output function for debugging purpose
		self.show_scores = theano.function(inputs=[self.inputPL, self.inputPR, self.inputNL, self.inputNR],
										   outputs=[self.scoreP, self.scoreN])
		self.show_hiddens = theano.function(inputs=[self.inputPL, self.inputPR, self.inputNL, self.inputNR],
											outputs=[self.hiddenP, self.hiddenN])
		if verbose:
			logger.debug('Architecture of {} built finished, summarized below: '.format(self.__class__.__name__))
			logger.debug('Input dimension: %d' % config.num_input)
			logger.debug('Hidden dimension of RNN: %d' % config.num_hidden)
			logger.debug('Hidden dimension of MLP: %d' % config.num_mlp)
			logger.debug('There are 2 BRNNEncoders used in the model.')
			logger.debug('Total number of parameters in this model: %d' % self.num_params)

	def update_params(self, grads, learn_rate):
		'''
		@grads: [np.ndarray]. List of numpy.ndarray for updating the model parameters.
				They are the corresponding gradients of model parameters.
		@learn_rate: scalar. Learning rate.
		'''
		for param, grad in zip(self.params, grads):
			p = param.get_value(borrow=True)
			param.set_value(p - learn_rate * grad, borrow=True)

	def set_params(self, params):
		'''
		@params: [np.ndarray]. List of numpy.ndarray to set the model parameters.
		'''
		for p, param in zip(self.params, params):
			p.set_value(param, borrow=True)

	def deepcopy(self, brnn):
		'''
		@brnn: BRNNMatchScorer. Copy the model parameters of another BRNNMatchScorer.
		'''
		assert len(self.params) == len(brnn.params)
		for p, param in zip(self.params, brnn.params):
			val = param.get_value()
			p.set_value(val)

	@staticmethod
	def save(fname, model):
		'''
		@fname: String. Filename to store the model.
		@model: BRNNMatcher. An instance of BRNNMatcher to be saved.
		'''
		with file(fname, 'wb') as fout:
			cPickle.dump(model, fout)

	@staticmethod
	def load(fname):
		'''
		@fname: String. Filename to load the model.
		'''
		with file(fname, 'rb') as fin:
			model = cPickle.load(fin)
		return model


